import pygame

from pyscail.settings import Settings
from pyscail.graphics import Graphics
from pyscail.grid import Grid


class Scail:
    grid: Grid | None = None


def mainLoop(initialize, update, settings: Settings):
    """Initializes all the classes and then updates them in a loop until the game is over"""

    settings.current_fps = settings.unpaused_fps
    grid = Grid(settings.game_width, settings.game_height, initialize, settings.mutate)
    Scail.grid = grid
    clock = pygame.time.Clock()
    graphics = Graphics(settings)

    gameExit = False
    while not gameExit:
        clock.tick(settings.current_fps)

        graphics.render_grid(grid)
        if not settings.paused:
            grid.step_grid()
            update()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                gameExit = True
                break


def run(initialize, update, settings: Settings):
    """Runs the specified cellular automaton

    initialize - a function to initialize the grid,
    with parameters i, j, width, height

    update - a function that is run every generation. No parameters
    """

    pygame.init()
    pygame.display.set_caption("SCAIL")
    mainLoop(initialize, update, settings)
    pygame.quit()


def cells(indices):

    return [Scail.grid.cells[i][j] for (i, j) in indices]



"""A bunch of utility functions for getting neighborhoods"""


def no_wrap_moore_neighborhood(i, j, width, height):
    result = []
    for x in range(max(i - 1, 0), min(i + 2, width)):
        for y in range(max(j - 1, 0), min(j + 2, height)):
            if (x, y) != (i, j):
                result.append((x, y))

    return result


def wrap_moore_neighborhood(i, j, width, height):
    result = []
    for x in range(i - 1, i + 2):
        for y in range(j - 1, j + 2):
            if (x, y) != (i, j):
                result.append((x % width, y % height))

    return result



"""Langton's ant"""

from dataclasses import dataclass
from enum import Enum

import pyscail as scail


class BW(Enum):
    BLACK = 0
    WHITE = 1


class Direction(Enum):
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"

    @staticmethod
    def clockwise(direction):
        match (direction):
            case Direction.UP:
                return Direction.RIGHT
            case Direction.DOWN:
                return Direction.LEFT
            case Direction.LEFT:
                return Direction.UP
            case Direction.RIGHT:
                return Direction.DOWN

    @staticmethod
    def anti_clockwise(direction):
        match (direction):
            case Direction.UP:
                return Direction.LEFT
            case Direction.DOWN:
                return Direction.RIGHT
            case Direction.LEFT:
                return Direction.DOWN
            case Direction.RIGHT:
                return Direction.UP

    @staticmethod
    def forward_direction(direction):
        match (direction):
            case Direction.UP:
                return (0, -1)
            case Direction.DOWN:
                return (0, 1)
            case Direction.LEFT:
                return (-1, 0)
            case Direction.RIGHT:
                return (1, 0)


class Constants:
    SIMULATION_WIDTH = 200
    SIMULATION_HEIGHT = 150
    CELL_SIZE = 4
    INITIAL_FPS = 256

    ANT_START_X = 100
    ANT_START_Y = 75
    ANT_START_DIRECTION = Direction.RIGHT


class Ant:
    x: int = Constants.ANT_START_X
    y: int = Constants.ANT_START_Y
    facing: Direction = Constants.ANT_START_DIRECTION

    @classmethod
    def move(cls):
        cell = scail.cells([(cls.x, cls.y)])[0]
        if cell.color == BW.WHITE:
            Ant.facing = Direction.clockwise(Ant.facing)
            cell.color = BW.BLACK
            (dx, dy) = Direction.forward_direction(Ant.facing)
            Ant.x += dx
            Ant.y += dy
        else:
            Ant.facing = Direction.anti_clockwise(Ant.facing)
            cell.color = BW.WHITE
            (dx, dy) = Direction.forward_direction(Ant.facing)
            Ant.x += dx
            Ant.y += dy

        Ant.x %= Constants.SIMULATION_WIDTH
        Ant.y %= Constants.SIMULATION_HEIGHT


settings = (
    scail.Settings.default()
    .set_dimensions(
        Constants.SIMULATION_WIDTH, Constants.SIMULATION_HEIGHT, Constants.CELL_SIZE
    )
    .set_initial_fps(Constants.INITIAL_FPS)
    .set_mutate(True)
)


@dataclass
class Cell:
    i: int
    j: int
    color: BW

    def next(self):
        pass

    def display(self):
        if (self.i, self.j) == (Ant.x, Ant.y):
            return (255, 0, 0)

        color_val = 255 if self.color == BW.WHITE else 0
        return (color_val, color_val, color_val)


def initialize(i, j, width, height):
    return Cell(i, j, BW.WHITE)


if __name__ == "__main__":
    scail.run(initialize, Ant.move, settings)



from typing import Any, List


class Grid:
    def __init__(self, width, height, initialize, mutate=True):
        self.width = width
        self.height = height
        self.initialize = initialize
        self.cells: List[List[Any]] = self.initial_cells()
        self.mutate = mutate

    def step_grid(self):
        if self.mutate:
            for i in range(self.width):
                for j in range(self.height):
                    self.cells[i][j].next()
        else:
            new_cells = [[None for _ in range(self.height)] for _ in range(self.width)]
            for i in range(self.width):
                for j in range(self.height):
                    new_cells[i][j] = self.cells[i][j].next()

            self.cells = new_cells

    def reset(self):
        self.cells = self.initial_cells()

    def initial_cells(self):
        cells = [
            [self.initialize(i, j, self.width, self.height) for j in range(self.height)]
            for i in range(self.width)
        ]

        return cells



"""Unit tests for the scail library, currently all in one file"""

import random

import kernels


def main():
    test_log(0, "beginning tests")
    test_kernels(100)

    test_log(0, "All tests passed successfully!")


def test_kernels(num_tests: int):
    test_log(1, "testing kernels")
    WIDTH = 100
    HEIGHT = 200
    for _ in range(num_tests):
        x = int(random.random() * WIDTH)
        y = int(random.random() * HEIGHT)

        test_wrap_moore(x, y, WIDTH, HEIGHT)
        test_nowrap_moore(x, y, WIDTH, HEIGHT)

    test_log(1, "Neighborhood kernel tests successfully passed")


def test_wrap_moore(x: int, y: int, width: int, height: int):
    neighbors = kernels.wrap_moore_neighborhood(x, y, width, height)
    assert len(neighbors) == 8

    for neighbor in neighbors:
        nx, ny = neighbor
        assert (x, y) != (nx, ny)
        assert nx == x or nx == (x + 1) % width or nx == (x - 1) % width
        assert ny == y or ny == (y + 1) % height or ny == (y - 1) % height


def test_nowrap_moore(x: int, y: int, width: int, height: int):
    neighbors = kernels.no_wrap_moore_neighborhood(x, y, width, height)

    length = len(neighbors)
    if x == 0 or x == width - 1:
        if y == 0 or y == height - 1:
            assert length == 3
        else:
            assert length == 5
    elif y == 0 or y == height - 1:
        assert length == 5
    else:
        assert length == 8

    for neighbor in neighbors:
        assert (x, y) != neighbor
        assert abs(x - neighbor[0]) <= 1
        assert abs(y - neighbor[1]) <= 1


def test_log(level: int, message: str):
    tabs = "  " * level
    print(tabs + message)


if __name__ == "__main__":
    main()



from dataclasses import dataclass

import pyscail as scail


class Constants:
    SIMULATION_WIDTH = 200
    SIMULATION_HEIGHT = 150
    CELL_SIZE = 4
    INITIAL_FPS = 8

    R_PENTOMINO = [(100, 100), (100, 101), (100, 99), (99, 100), (101, 99)]


settings = (
    scail.Settings.default()
    .set_dimensions(
        Constants.SIMULATION_WIDTH, Constants.SIMULATION_HEIGHT, Constants.CELL_SIZE
    )
    .set_initial_fps(Constants.INITIAL_FPS)
)


@dataclass
class Cell:
    i: int
    j: int
    alive: bool

    def next(self):
        neighbors = scail.cells(
            scail.kernels.wrap_moore_neighborhood(
                self.i,
                self.j,
                Constants.SIMULATION_WIDTH,
                Constants.SIMULATION_HEIGHT,
            )
        )

        num_alive = 0
        for cell in neighbors:
            if cell.alive:
                num_alive += 1

        if num_alive < 2 or num_alive > 3:
            alive_in_next_gen = False
        elif num_alive == 3:
            alive_in_next_gen = True
        else:
            alive_in_next_gen = self.alive

        return Cell(self.i, self.j, alive_in_next_gen)

    def display(self):
        color_val = 255 if self.alive else 0
        return (color_val, color_val, color_val)


def initialize(i, j, width, height):
    # r-pentomino
    init_list = Constants.R_PENTOMINO
    alive = True if (i, j) in init_list else False
    return Cell(i, j, alive)


if __name__ == "__main__":
    scail.run(initialize, lambda: None, settings)



from dataclasses import dataclass
from math import floor
import random

import pyscail as scail


class Constants:
    SIMULATION_WIDTH = 400
    SIMULATION_HEIGHT = 300
    CELL_SIZE = 1
    INITIAL_FPS = 8

    K1 = 2
    K2 = 3
    NUM_STATES = 200
    G = 70
    NUM_COLORS = 8
    COLORS = [(i, 0, i) for i in range(0, 256, int(256 / NUM_COLORS))]


settings = (
    scail.Settings.default()
    .set_dimensions(
        Constants.SIMULATION_WIDTH, Constants.SIMULATION_HEIGHT, Constants.CELL_SIZE
    )
    .set_initial_fps(Constants.INITIAL_FPS)
)


@dataclass
class Cell:
    i: int
    j: int
    state: int

    def next(self):
        neighbors = scail.cells(
            scail.kernels.wrap_moore_neighborhood(
                self.i,
                self.j,
                Constants.SIMULATION_WIDTH,
                Constants.SIMULATION_HEIGHT,
            )
        )

        if self.state == Constants.NUM_STATES - 1:
            next_state = 0
        else:
            ill = 0
            infected = 0
            for neighbor in neighbors:
                if neighbor.state == Constants.NUM_STATES - 1:
                    ill += 1
                elif neighbor.state != 0:
                    infected += 1

            if self.state == 0:
                next_state = floor(infected / Constants.K1) + floor(ill / Constants.K2)
            else:
                s = self.state + sum([neighbor.state for neighbor in neighbors])
                next_state = floor(s / (1 + infected + ill)) + Constants.G

        if next_state > Constants.NUM_STATES - 1:
            next_state = Constants.NUM_STATES - 1

        return Cell(self.i, self.j, next_state)

    def display(self):
        index = floor(Constants.NUM_COLORS * self.state / Constants.NUM_STATES)
        return Constants.COLORS[index]


def initialize(i, j, width, height):
    state = floor(random.random() * Constants.NUM_STATES)
    return Cell(i, j, state)


if __name__ == "__main__":
    scail.run(initialize, lambda: None, settings)



"""Provides the Graphics class, which draws everything on the screen"""

import pygame

from pyscail.settings import Settings


class Graphics:
    """Draws to and updates the screen"""

    def __init__(self, settings: Settings):
        self.__settings = settings
        self.__resolution_width = self.__settings.game_width * self.__settings.cell_size
        self.__resolution_height = (
            self.__settings.game_height * self.__settings.cell_size
        )
        self.__screen = pygame.display.set_mode(
            (self.__resolution_width, self.__resolution_height)
        )

    def render_grid(self, grid):
        """Draws the grid of cells onto the screen"""
        self.__screen.fill((0, 0, 0))
        for i in range(grid.width):
            for j in range(grid.height):
                self.__render_cell(i, j, grid.cells[i][j])

        # Update the screen so that the changes take place
        pygame.display.update()

    def __render_cell(self, i: int, j: int, cell):
        """Renders a cell at index i,j

        Its actual position depends on the resolution, which depends on the cell size
        So a cell of index (i,j) is rendered at (top corner) = (i*cell_size, j*cell_size)
        """

        color = cell.display()
        size = self.__settings.cell_size
        rect = (i * size, j * size, size, size)
        pygame.draw.rect(self.__screen, color, rect)



from dataclasses import dataclass

import pyscail as scail


class Constants:
    REAL_LEFT_BOUND = -2
    REAL_RIGHT_BOUND = 1.5
    IMAGINARY_UP_BOUND = 1.5
    IMAGINARY_DOWN_BOUND = -2

    MAGNITUDE_CEILING = 2

    JULIA_CENTER = complex(0, 0.8)

    WIDTH = 800
    HEIGHT = 800
    CELL_SIZE = 1


class Globals:
    num_generations: int = 1

    @classmethod
    def update(cls):
        cls.num_generations += 1


settings = (
    scail.Settings.default()
    .set_dimensions(Constants.WIDTH, Constants.HEIGHT, Constants.CELL_SIZE)
    .set_mutate(True)
)


@dataclass
class State:
    z: complex
    lost: bool = False
    num_generation_alive: int = 0

    def next(self):
        """z(n+1) = z(n)^2 + c"""
        if self.lost:
            return

        self.num_generation_alive += 1
        self.lost = abs(self.z) > Constants.MAGNITUDE_CEILING
        self.z = self.z**2 + Constants.JULIA_CENTER

    def display(self):
        color_val = int(
            lerp(255, 0, self.num_generation_alive / Globals.num_generations)
        )
        return (color_val, color_val, color_val)


def initialize(i, j, width, height):
    cr = lerp(
        Constants.REAL_LEFT_BOUND,
        Constants.REAL_RIGHT_BOUND,
        i / width,
    )

    ci = lerp(
        Constants.IMAGINARY_DOWN_BOUND,
        Constants.IMAGINARY_UP_BOUND,
        j / height,
    )

    return State(complex(cr, ci))


def lerp(a, b, t):
    return a + (b - a) * t


if __name__ == "__main__":
    scail.run(initialize, Globals.update, settings)



"""Provides Constants and Settings, some classes to hold data

Constants holds config data that doesn't change during execution
Settings holds data that needs to be shared between other classes and relates to the game settings and state
"""

from dataclasses import dataclass
from typing import Self


class Defaults:
    GAME_WIDTH = 200  # Specified
    GAME_HEIGHT = 150  # Specified
    CELL_SIZE = 4  # optional
    INITIAL_FPS = 8  # optional
    MAX_FPS = 128  # optional
    MIN_FPS = 1  # optional
    MUTATE = False
    PAUSED = False


# Maybe the scail-specific constants can go in globals, or just a "Constants" class
#
# For the Settings class, the user can maybe define another settings class, and we can do a
# hasattr analysis


# None of these are scail-specific, so they will come from the "Settings constants"
@dataclass
class Settings:
    game_width: int
    game_height: int
    cell_size: int
    unpaused_fps: int  # The fps when the game is not paused
    # The actual fps at which the screen updates. Needs to be separated from unpaused fps because
    # if fps is changed when the game is paused, its effects should take place after it is unpaused
    current_fps: int
    paused: bool
    mutate: bool  # whether the autamaton does check-and-update or mutation

    @staticmethod
    def default():
        return Settings(
            Defaults.GAME_WIDTH,
            Defaults.GAME_HEIGHT,
            Defaults.CELL_SIZE,
            Defaults.INITIAL_FPS,
            Defaults.INITIAL_FPS,
            paused=Defaults.PAUSED,
            mutate=Defaults.MUTATE,
        )

    def set_dimensions(self, width: int, height: int, cell_size: int) -> Self:
        """Sets the dimensions of the program and returns the mutated object

        The resolution will be (width * cell_size) x (height * cell_size)
        """
        self.game_width = width
        self.game_height = height
        self.cell_size = cell_size
        return self

    def set_initial_fps(self, fps: int):
        """Sets the initial fps of the program and returns the mutated object

        FPS must be between 1 and 128, inclusive
        """

        self.current_fps = self.unpaused_fps = fps
        return self

    def set_mutate(self, mutate: bool):
        """Sets the mutate of the scail and returns the mutated object"""
        self.mutate = mutate
        return self

    def __str__(self) -> str:
        return f"Settings(width: {self.game_width}, height: {self.game_height}, cell_size = {self.cell_size}, ...)"



"""The wireworld cellular automaton, on a random map"""

from enum import Enum
import random

import pyscail as scail


class Constants:
    WIDTH = 200
    HEIGHT = 150
    CELL_SIZE = 4
    INITIAL_FPS = 8

    WIRE_EPICENTER_DENSITY = 0.001
    WIRE_DENSITY = 0.1

    ELECTRON_DENSITY = 0.0005


settings = (
    scail.Settings.default()
    .set_dimensions(Constants.WIDTH, Constants.HEIGHT, Constants.CELL_SIZE)
    .set_mutate(False)
)


class Colors:
    red = (255, 0, 0)
    blue = (0, 0, 255)
    yellow = (255, 255, 0)
    black = (0, 0, 0)


class State(Enum):
    EMPTY = "empty"
    CONDUCTOR = "conductor"
    ELECTRON_HEAD = "electron head"
    ELECTRON_TAIL = "electron tail"


def make_wireworld(width: int, height: int):
    map = [[State.EMPTY for _ in range(height)] for _ in range(width)]
    num_epicenters = int(Constants.WIRE_EPICENTER_DENSITY * width * height)
    epicenters = [random_point(width, height) for _ in range(num_epicenters)]

    epicenter_pairs = []
    for i in range(num_epicenters):
        for j in range(i + 1, num_epicenters):
            epicenter_pairs.append((epicenters[i], epicenters[j]))

    wires_to_spawn = width * height * Constants.WIRE_DENSITY
    while wires_to_spawn > 0 and len(epicenter_pairs) > 0:
        pair = epicenter_pairs[int(random.random() * len(epicenter_pairs))]
        wires_expended = join_pair(pair, map)
        wires_to_spawn -= wires_expended
        epicenter_pairs.remove(pair)

    return map


def random_point(width, height):
    rx = int(random.random() * width)
    ry = int(random.random() * height)
    return (rx, ry)


def join_pair(pair, map):
    ((x1, y1), (x2, y2)) = pair
    map[x1][y1] = State.CONDUCTOR
    map[x2][y2] = State.CONDUCTOR

    minx, maxx = min(x1, x2), max(x1, x2)
    miny, maxy = min(y1, y2), max(y1, y2)

    wires_laid = 0

    for i in range(minx, maxx + 1):
        map[i][miny] = State.CONDUCTOR
        map[i][maxy] = State.CONDUCTOR
        wires_laid += 2

    for j in range(miny, maxy + 1):
        map[minx][j] = State.CONDUCTOR
        map[maxx][j] = State.CONDUCTOR
        wires_laid += 2

    return wires_laid


class Cell:
    transmission_map = {
        State.EMPTY: State.EMPTY,
        State.ELECTRON_HEAD: State.ELECTRON_TAIL,
        State.ELECTRON_TAIL: State.CONDUCTOR,
    }

    def __init__(self, i: int, j: int, state: State):
        self.i = i
        self.j = j
        self.state = state

    def next(self):
        if self.state != State.CONDUCTOR:
            next_state = Cell.transmission_map[self.state]
        else:
            num_neighboring_electrons = 0
            neighbors = scail.cells(
                scail.kernels.wrap_moore_neighborhood(
                    self.i, self.j, Constants.WIDTH, Constants.HEIGHT
                )
            )
            for neighbor in neighbors:
                if neighbor.state == State.ELECTRON_HEAD:
                    num_neighboring_electrons += 1

            if num_neighboring_electrons == 1 or num_neighboring_electrons == 2:
                next_state = State.ELECTRON_HEAD
            else:
                next_state = State.CONDUCTOR

        return Cell(self.i, self.j, next_state)

    def display(self):
        match (self.state):
            case State.EMPTY:
                return Colors.black
            case State.CONDUCTOR:
                return Colors.yellow
            case State.ELECTRON_HEAD:
                return Colors.blue
            case State.ELECTRON_TAIL:
                return Colors.red


map = make_wireworld(Constants.WIDTH, Constants.HEIGHT)


def initialize(i: int, j: int, width: int, height: int):
    state = map[i][j]
    if state == State.CONDUCTOR and random.random() < Constants.ELECTRON_DENSITY:
        state = State.ELECTRON_HEAD
    return Cell(i, j, state)


if __name__ == "__main__":
    scail.run(initialize, lambda: None, settings)
